@@ -359,6 +359,10 @@ build_ssdt(void)
     else
         ssdt_ptr[acpi_s4_pkg[0] + 1] = ssdt[acpi_s4_pkg[0] + 3] = sys_states[4] & 127;
 
+    dprintf(1, "pcimem 0x%llx-0x%llx 0x%llx-0x%llx\n",
+	    pcimem_start, pcimem_end,
+	    pcimem64_start, pcimem64_end);
+
     // store pci io windows
     *(u32*)&ssdt_ptr[acpi_pci32_start[0]] = cpu_to_le32(pcimem_start);
     *(u32*)&ssdt_ptr[acpi_pci32_end[0]] = cpu_to_le32(pcimem_end - 1);
