@@ -597,9 +597,7 @@ ahci_controller_setup(struct pci_device *pci)
     struct ahci_port_s *port;
     u16 bdf = pci->bdf;
     u32 val, pnr, max;
-#ifdef AHCI_HBA_RESET
     u64 end;
-#endif
 
     if (!ctrl) {
         warn_noalloc();
@@ -624,20 +622,21 @@ ahci_controller_setup(struct pci_device *pci)
     pci_config_maskw(bdf, PCI_COMMAND, 0,
                      PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);
 
-#if AHCI_HBA_RESET
-    /* HBA Reset */
-    val = ahci_ctrl_readl(ctrl, HOST_CTL);
-    ahci_ctrl_writel(ctrl, HOST_CTL, val | HOST_CTL_AHCI_EN |
-                     HOST_CTL_RESET);
-    end = calc_future_tsc(AHCI_HBA_RESET_TIMEOUT);
-    while (ahci_ctrl_readl(ctrl, HOST_CTL) & HOST_CTL_AHCI_EN) {
-        if (check_tsc(end)) {
-            warn_timeout();
-            return;
+    if (CONFIG_AHCI_HBA_RESET) {
+        /* HBA Reset */
+        dprintf(1, "AHCI HBA Reset\n");
+        val = ahci_ctrl_readl(ctrl, HOST_CTL);
+        ahci_ctrl_writel(ctrl, HOST_CTL, val | HOST_CTL_AHCI_EN |
+                         HOST_CTL_RESET);
+        end = calc_future_tsc(AHCI_HBA_RESET_TIMEOUT);
+        while (ahci_ctrl_readl(ctrl, HOST_CTL) & HOST_CTL_RESET) {
+            if (check_tsc(end)) {
+                warn_timeout();
+                return;
+            }
+            yield();
         }
-        yield();
     }
-#endif
 
     /* Enable AHCI mode */
     val = ahci_ctrl_readl(ctrl, HOST_CTL);
