@@ -532,6 +532,51 @@ pci_bios_get_bar(struct pci_device *pci, int bar,
     *pis64 = is64;
 }
 
+static void
+pci_bios_get_bar_val(struct pci_device *pci, int bar,
+		     int *ptype, u64 *addr, u64 *psize, int *pis64)
+{
+    u32 ofs = pci_bar(pci, bar);
+    u16 bdf = pci->bdf;
+    u32 old = pci_config_readl(bdf, ofs);
+    int is64 = 0, type = PCI_REGION_TYPE_MEM;
+    u64 mask;
+
+    if (bar == PCI_ROM_SLOT) {
+        mask = PCI_ROM_ADDRESS_MASK;
+        pci_config_writel(bdf, ofs, mask);
+    } else {
+        if (old & PCI_BASE_ADDRESS_SPACE_IO) {
+            mask = PCI_BASE_ADDRESS_IO_MASK;
+            type = PCI_REGION_TYPE_IO;
+        } else {
+            mask = PCI_BASE_ADDRESS_MEM_MASK;
+            if (old & PCI_BASE_ADDRESS_MEM_PREFETCH)
+                type = PCI_REGION_TYPE_PREFMEM;
+            is64 = ((old & PCI_BASE_ADDRESS_MEM_TYPE_MASK)
+                    == PCI_BASE_ADDRESS_MEM_TYPE_64);
+        }
+        pci_config_writel(bdf, ofs, ~0);
+    }
+    u64 val = pci_config_readl(bdf, ofs);
+    pci_config_writel(bdf, ofs, old);
+    if (is64) {
+        u32 hold = pci_config_readl(bdf, ofs + 4);
+        pci_config_writel(bdf, ofs + 4, ~0);
+        u32 high = pci_config_readl(bdf, ofs + 4);
+        pci_config_writel(bdf, ofs + 4, hold);
+        val |= ((u64)high << 32);
+        mask |= ((u64)0xffffffff << 32);
+        *addr = (((u64)hold << 32) | old) & mask;
+        *psize = (~(val & mask)) + 1;
+    } else {
+	*addr = old & mask;
+        *psize = ((~(val & mask)) + 1) & 0xffffffff;
+    }
+    *ptype = type;
+    *pis64 = is64;
+}
+
 static int pci_bios_bridge_region_is64(struct pci_region *r,
                                  struct pci_device *pci, int type)
 {
@@ -678,6 +723,92 @@ static int pci_bios_check_devices(struct pci_bus *busses)
     return 0;
 }
 
+static void pci_bios_update_start_end(u64 start, u64 end)
+{
+    if (end <= 0x100000000) {
+        if (pcimem_start > start) {
+            pcimem_start = start;
+        }
+        if (pcimem_end < end) {
+            pcimem_end = end;
+        }
+    } else if (start >= 0x100000000) {
+        if (pcimem64_start > start) {
+            pcimem64_start = start;
+        }
+        if (pcimem64_end < end) {
+            pcimem64_end = end;
+        }
+    }
+}
+
+static int pci_bios_check_start_end(void)
+{
+    pcimem_start = 0xffffffff;
+    pcimem_end = 0x0;
+    pcimem64_start = 0xffffffffffffffff;
+    pcimem64_end = 0x0;
+
+    struct pci_device *pci;
+    foreachpci(pci) {
+        int i;
+        u16 bdf = pci->bdf;
+        for (i = 0; i < PCI_NUM_REGIONS; i++) {
+            if ((pci->class == PCI_CLASS_BRIDGE_PCI) &&
+                (i >= PCI_BRIDGE_NUM_REGIONS && i < PCI_ROM_SLOT))
+                continue;
+            int type, is64;
+            u64 addr, size;
+            pci_bios_get_bar_val(pci, i, &type, &addr, &size, &is64);
+            if (size == 0)
+                continue;
+
+            if (type == PCI_REGION_TYPE_IO) {
+                continue;
+            }
+            if (size < PCI_DEVICE_MEM_MIN)
+                size = PCI_DEVICE_MEM_MIN;
+            pci_bios_update_start_end(addr, addr + size);
+            if (is64)
+                i++;
+        }
+
+        if (pci->class != PCI_CLASS_BRIDGE_PCI) {
+            continue;
+        }
+        u64 addr, limit, val;
+
+        /* memory */
+        addr = (u64)(pci_config_readw(bdf, PCI_MEMORY_BASE) & 0xfff0) << 16;
+        limit = ((u64)(pci_config_readw(bdf, PCI_MEMORY_LIMIT) & 0xfff0) << 16)
+            + 0xfffff;
+        if (addr <= limit) {
+            pci_bios_update_start_end(addr, limit + 1);
+        }
+
+        /* prefetchable memory */
+        val = pci_config_readw(bdf, PCI_PREF_MEMORY_BASE);
+        addr = (val & 0xfff0) << 16;
+        if (val & 0x1) {
+            addr |= (u64)pci_config_readl(bdf, PCI_PREF_BASE_UPPER32) << 32;
+        }
+        val = pci_config_readw(bdf, PCI_PREF_MEMORY_LIMIT);
+        limit = ((val & 0xfff0) << 16) + 0xfffff;;
+        if (val & 0x1) {
+            limit |= (u64)pci_config_readl(bdf, PCI_PREF_LIMIT_UPPER32) << 32;
+        }
+        if (addr <= limit) {
+            pci_bios_update_start_end(addr, limit + 1);
+        }
+    }
+    if (pcimem64_start == 0xffffffffffffffff) {
+        pcimem64_start = 0;
+    }
+
+    dprintf(1, "pcimem %08llx-%08llx pcimem64 %08llx-%08llx\n",
+            pcimem_start, pcimem_end, pcimem64_start, pcimem64_end);
+    return 0;
+}
 
 /****************************************************************
  * BAR assignment
@@ -815,6 +946,7 @@ pci_setup(void)
 
     if (CONFIG_NO_PCI_SETUP) {
         pci_probe_devices();
+        pci_bios_check_start_end();
         return;
     }
 
