@@ -27,15 +27,16 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 /*
- * Copyright (c) 2010-2012 Yuichi Watanabe
+ * Copyright (c) 2010-2014 Yuichi Watanabe
  */
 
 #include <core/acpi.h>
 #include <core/assert.h>
 #include <core/initfunc.h>
 #include <core/printf.h>
+#include <core/sleep.h>
 #include <core/string.h>
-
+#include <core/vm.h>
 #include "acpi.h"
 #include "asm.h"
 #include "acpi_dsdt.h"
@@ -44,7 +45,7 @@
 #include "io_io.h"
 #include "mm.h"
 #include "panic.h"
-#include <core/sleep.h>
+#include "vm.h"
 
 #define FIND_RSDP_NOT_FOUND	0xFFFFFFFFFFFFFFFFULL
 #define RSDP_SIGNATURE		"RSD PTR"
@@ -63,6 +64,18 @@
 
 #define PARSERS_MAX_NUM		8
 
+#define ACPI_FACP_FLAG_RESET_REG_SUP	0x400
+
+/* #define ACPI_DEBUG */
+#ifdef ACPI_DEBUG
+#define ACPI_DBG(...)						\
+	do {							\
+		printf("ACPI: " __VA_ARGS__);			\
+	} while (0)
+#else
+#define ACPI_DBG(...)
+#endif
+
 struct acpi_parser {
 	char signature[SIGNATURE_LEN];
 	acpi_parser_t *func;
@@ -129,7 +142,7 @@ struct facp {
 	u16 iapc_boot_arch;
 	u8 reserved2;
 	u32 flags;
-	u8 reset_reg[12];
+	struct gas reset_reg;
 	u8 reset_value;
 	u8 reserved3[3];
 	u64 x_firmware_ctrl;
@@ -156,15 +169,18 @@ struct facs {
 	u8 reserved[31];
 } __attribute__ ((packed));
 
-static bool rsdp_found;
+static bool rsdp_found = false;
 static struct rsdp rsdp_copy;
 
 static struct acpi_parser parsers[PARSERS_MAX_NUM];
 static uint parsers_num = 0;
 
-static bool pm1a_cnt_found;
+static bool pm1a_cnt_found = false;
 static ioport_t pm1a_cnt_ioaddr;
 static u64 facs_addr;
+static bool reset_reg_found = false;
+static ioport_t reset_reg_ioaddr;
+static u8 reset_value;
 
 unsigned char acpi_dsdt_system_state[6][5];
 
@@ -179,6 +195,22 @@ acpi_checksum (void *p, int len)
 	return s;
 }
 
+void
+acpi_update_checksum(struct acpi_desc_header *header)
+{
+	u8 sum;
+
+	sum = acpi_checksum(header, header->length);
+	header->checksum -= sum;
+}
+
+void
+acpi_hide_table(struct acpi_desc_header *header)
+{
+	memset(header->signature, 0, 4);
+	acpi_update_checksum(header);
+}
+
 static void *
 acpi_mapmem (u64 addr, int len)
 {
@@ -344,6 +376,7 @@ acpi_poweroff (void)
 		return;
 	if (!acpi_dsdt_system_state[5][0])
 		return;
+	ACPI_DBG("acpi_poweroff\n");
 	typx = acpi_dsdt_system_state[5][1] << PM1_CNT_SLP_TYPX_SHIFT;
 	/* FIXME: how to handle pm1b_cnt? */
 	asm_inl (pm1a_cnt_ioaddr, &data);
@@ -363,6 +396,8 @@ acpi_pm1_sleep (u32 v)
 	u8 n;
 #endif
 
+	ACPI_DBG("acpi_pm1_sleep 0x%x\n", v);
+
 #ifdef ACPI_DSDT
 	n = (v & PM1_CNT_SLP_TYPX_MASK) >> PM1_CNT_SLP_TYPX_SHIFT;
 	for (i = 0; i <= 5; i++) {
@@ -372,8 +407,10 @@ acpi_pm1_sleep (u32 v)
 		else
 			m[i] = false;
 	}
-	if (!m[2] && !m[3])
+	if (!m[2] && !m[3]) {
+		ACPI_DBG("acpi_pm1_sleep m[2] %d m[3] %d\n", m[2], m[3]);
 		return;
+	}
 #endif
 	facs = acpi_mapmem (facs_addr, sizeof *facs);
 	if (IS_STRUCT_SIZE_OK (facs->length, facs,
@@ -434,11 +471,49 @@ acpi_io_monitor (iotype_t type, ioport_t port, void *data)
 	return 0; /* pass */
 }
 
+static int
+acpi_reset_reg (iotype_t type, ioport_t port, void *data)
+{
+	u32 v;
+
+	switch (type) {
+	case IOTYPE_OUTB:
+		v = *(u8 *)data;
+		break;
+	case IOTYPE_OUTW:
+		v = *(u16 *)data;
+		break;
+	case IOTYPE_OUTL:
+		v = *(u32 *)data;
+		break;
+	default:
+		return 0; /* pass */
+	}
+
+	if (v == reset_value) {
+		if (vm_reset() == VMMERR_SUCCESS) {
+			return 1; /* emulated */
+		}
+		printf("ACPI reset of %s ignored\n",
+		       vm_get_name());
+		return 1; /* emulated */
+	}
+
+	return 0; /* pass */
+}
+
 static void
 acpi_iohook (void)
 {
+	if (vm_get_id() != 0) {
+		return;
+	}
+
 	if (pm1a_cnt_found)
 		set_iofunc (pm1a_cnt_ioaddr, acpi_io_monitor);
+
+	if (reset_reg_found && reset_reg_ioaddr != 0x64)
+		set_iofunc(reset_reg_ioaddr, acpi_reset_reg);
 }
 
 static void
@@ -460,6 +535,35 @@ get_pm1a_cnt_ioaddr (struct facp *q)
 	} else {
 		panic ("ACPI FACP is too short");
 	}
+	pm1a_cnt_found = true;
+	ACPI_DBG("PM1a control port is 0x%X\n", pm1a_cnt_ioaddr);
+}
+
+static void
+get_reset_reg_ioaddr (struct facp *facp)
+{
+	if ((facp->flags & ACPI_FACP_FLAG_RESET_REG_SUP) == 0) {
+		printf("ACPI reset register is not supported.\n");
+		return;
+	}
+	if (!IS_STRUCT_SIZE_OK(facp->header.length, facp, facp->reset_value)) {
+		printf("FACP length is too short: %d\n", facp->header.length);
+		return;
+	}
+	if (facp->reset_reg.address_space_id != ADDRESS_SPACE_ID_IO) {
+		printf("ACPI RESET_REG is not in I/O space: %d\n",
+		       facp->reset_reg.address_space_id);
+		return;
+	}
+	if (facp->reset_reg.address > 0xFFFF) {
+		printf("ACPI RESET_REG is out of range: 0x%llx\n",
+		       facp->reset_reg.address);
+		return;
+	}
+	reset_reg_found = true;
+	reset_reg_ioaddr = facp->reset_reg.address;
+	reset_value = facp->reset_value;
+	ACPI_DBG("RESET_REG 0x%x value 0x%x\n", reset_reg_ioaddr, reset_value);
 }
 
 static void
@@ -485,12 +589,10 @@ parse_facp(void *table, u32 length)
 	acpi_dsdt_parse(facp->dsdt);
 #endif
 	get_pm1a_cnt_ioaddr(facp);
+	get_reset_reg_ioaddr(facp);
 	get_facs_addr(facp);
 	if (0)
 		debug_dump(facp, facp->header.length);
-	if (0)
-		printf("PM1a control port is 0x%X\n", pm1a_cnt_ioaddr);
-	pm1a_cnt_found = true;
 }
 
 static void
@@ -499,9 +601,6 @@ acpi_init_global (void)
 	u64 rsdp;
 	struct rsdp *p;
 
-	rsdp_found = false;
-	pm1a_cnt_found = false;
-
 	acpi_register_parser(FACP_SIGNATURE, parse_facp);
 
 	rsdp = find_rsdp ();
@@ -517,4 +616,4 @@ acpi_init_global (void)
 }
 
 INITFUNC ("global3", acpi_init_global);
-INITFUNC ("passvm1", acpi_iohook);
+INITFUNC ("setupvm1", acpi_iohook);
