@@ -27,9 +27,10 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 /*
- * Copyright (c) 2010-2012 Yuichi Watanabe
+ * Copyright (c) 2010-2013 Yuichi Watanabe
  */
 
+#include <core/assert.h>
 #include <core/cpu.h>
 #include <core/initfunc.h>
 #include <core/mm.h>
@@ -46,12 +47,6 @@
 #include "panic.h"
 #include "vcpu.h"
 
-/* #define MWAIT_DEBUG */
-#ifdef MWAIT_DEBUG
-#include <core/printf.h>
-#include <core/time.h>
-#endif
-
 /* #define APIC_DEBUG */
 #include <core/printf.h>
 #ifdef APIC_DEBUG
@@ -63,6 +58,8 @@
 #define APIC_DBG(...)
 #endif
 
+#define FAKE_SIPI_VECTOR 0x31
+
 void
 apic_read_cr8 (u64 *val)
 {
@@ -91,55 +88,69 @@ apic_base_changed (ulong apic_base)
 	}
 }
 
-static void
-apic_enter_wait_for_sipi_state_without_lock(void)
-{
-	current->mwait->sipi_pending = false;
-	current->apic_data.sipi_vector = 0;
-	current->apic_data.wait_for_sipi = true;
-}
-
 void
 apic_enter_wait_for_sipi_state(void)
 {
 	spinlock_lock(&current->apic_data.lock);
-	apic_enter_wait_for_sipi_state_without_lock();
+	current->apic_data.sipi_pending = false;
+	current->apic_data.sipi_vector = 0;
+	current->apic_data.wait_for_sipi = true;
 	spinlock_unlock(&current->apic_data.lock);
 }
 
+static int
+fake_sipi_handler(vector_t vector, void *data)
+{
+	int ret = 0;
+
+	spinlock_lock(&current->apic_data.lock);
+	if (current->apic_data.fake_sipi) {
+		current->apic_data.fake_sipi = false;
+		ret = 1;
+	}
+	spinlock_unlock(&current->apic_data.lock);
+
+	return ret;
+}
+
 void
 apic_handle_wait_for_sipi_state (void)
 {
-	struct mwait_data *mwait = current->mwait;
-#ifdef MWAIT_DEBUG
-	u64 delta;
-#endif
+	ASSERT(currentcpu->disable_interrupt == 1);
 
 	spinlock_lock(&current->apic_data.lock);
+	if (!current->apic_data.wait_for_sipi) {
+		spinlock_unlock(&current->apic_data.lock);
+		return;
+	}
+	
+	APIC_DBG("handle wait-for-sipi state\n");
+
 	while (current->apic_data.wait_for_sipi) {
-		setup_monitor(mwait);
-		if (!mwait->sipi_pending) {
+		if (!current->apic_data.sipi_pending) {
+			/* Keep interrupts disabled on unlocking */
 			spinlock_unlock(&current->apic_data.lock);
-			
-#ifdef MWAIT_DEBUG
-			delta = -get_cpu_time();
-#endif
-			wait_for_updating();
-#ifdef MWAIT_DEBUG
-			delta += get_cpu_time();
-			printf("DEBUG: cpu%d wait sipi delta %llu usec\n",
-			       get_cpu_id(), delta);
-#endif
+
+			/*
+			 * Enable interrupts and halt.
+			 * Please note execution of STI with RFLAGS.IF = 0
+			 * blocks interrupts for one instruction after
+			 * its execution.
+			 */
+			asm_sti_and_hlt();
+
 			spinlock_lock(&current->apic_data.lock);
 		}
-		if (mwait->sipi_pending) {
+		if (current->apic_data.sipi_pending != 0) {
 			current->vmctl.write_realmode_seg(SREG_CS,
 				  current->apic_data.sipi_vector << 8);
-			mwait->sipi_pending = false;
+			current->vmctl.write_ip(0);
+			current->apic_data.sipi_pending = false;
 			current->apic_data.sipi_vector = 0;
 			current->apic_data.wait_for_sipi = false;
 		}
 	}
+	APIC_DBG("Exit wait-for-sipi state\n");
 	spinlock_unlock(&current->apic_data.lock);
 }
 
@@ -166,13 +177,23 @@ call_send_init(struct vcpu *dest_vcpu, void *data)
 	return false;
 }
 
+void
+apic_send_init_others(void)
+{
+	vcpu_list_foreach(current->vm, call_send_init, NULL);
+}
+
 static void
 send_sipi (struct vcpu *dest_vcpu, vector_t vector)
 {
 	spinlock_lock(&dest_vcpu->apic_data.lock);
 	if (dest_vcpu->apic_data.wait_for_sipi) {
+		dest_vcpu->apic_data.sipi_pending = true;
 		dest_vcpu->apic_data.sipi_vector = vector;
-		dest_vcpu->mwait->sipi_pending = true;
+		if (!dest_vcpu->apic_data.fake_sipi) {
+			dest_vcpu->apic_data.fake_sipi = true;
+			apic_send_ipi(FAKE_SIPI_VECTOR, dest_vcpu->apic_id);
+		}
 	}
 	spinlock_unlock(&dest_vcpu->apic_data.lock);
 }
@@ -257,6 +278,7 @@ icr_low_handler(u32 *buf, bool wr)
 	case ICR_DELIVERY_MODE_LOWEST_PRI:
 		switch (dsh) {
 		case ICR_DSH_DEST:
+		case ICR_DSH_SELF:
 			disable_interrupt();
 			apic_wait_for_idle();
 			apic_write32(APIC_ICR_HIGH_OFFSET, icr_high);
@@ -297,8 +319,7 @@ icr_low_handler(u32 *buf, bool wr)
 			send_init(dest_vcpu);
 			break;
 		case ICR_DSH_OTHER:
-			vcpu_list_foreach(current->vm, call_send_init,
-					  NULL);
+			apic_send_init_others();
 			break;
 		default:
 			APIC_DBG("Specified DSH is not supported on INIT: CPU%d ICR %08x %08x\n",
@@ -356,7 +377,7 @@ eoi_handler(u32 *buf, bool wr)
 		/* Read */
 		return 0;
 	}
-	extint_apic_eoi();
+	apic_write_eoi();
 	return 1; /* emulated */
 }
 
@@ -443,12 +464,13 @@ static void
 apic_pass_init(void)
 {
 	spinlock_init(&current->apic_data.lock);
+	set_int_hook(FAKE_SIPI_VECTOR, fake_sipi_handler, NULL);
 }
 
 static void
 apic_initvm(void)
 {
-	if (cpu_is_bsp()) {
+	if (vm_get_id() == 0) {
 		mmio_register(APIC_BASE_ADDRESS, APIC_MEM_SIZE,
 			      apic_privileged_mmio_handler, NULL);
 	} else {
@@ -458,4 +480,4 @@ apic_initvm(void)
 }
 
 INITFUNC ("passcpu0", apic_pass_init);
-INITFUNC ("passvm1", apic_initvm);
+INITFUNC ("setupvm1", apic_initvm);
