@@ -34,14 +34,16 @@
 #include <core/mm.h>
 #include <core/param.h>
 #include <core/printf.h>
+#include <core/sleep.h>
 #include <core/spinlock.h>
 #include <io/pci.h>
+#include "apic.h"
 #include "apic_pass.h"
 #include "current.h"
 #include "gbiosloader.h"
 #include "int.h"
 #include "pcpu.h"
-#include <core/sleep.h>
+#include "setup_guestboot.h"
 #include "vm.h"
 
 static size_t driver_data_size = 0;
@@ -289,6 +291,8 @@ vm_new(void)
 	spinlock_unlock(&vm_list_lock);
 
 	vm->vbsp_apic_id = get_apic_id();
+
+	wait_init(&vm->wait);
 	return vm;
 }
 
@@ -326,33 +330,62 @@ vm_iommu_enabled(void)
 	current->vm->iommu_enabled = true;
 }
 
+void
+vm_wait_other_cpus(void)
+{
+	wait(&current->vm->wait, vm_get_vcpu_count());
+}
+
+bool
+vm_resetting(void)
+{
+	return current->vm->resetting;
+}
+
 vmmerr_t
 vm_reset(void)
 {
-	if (current->vm->gmm.clear_mem == NULL ||
-	    !current->vbsp) {
+	struct vm *vm = current->vm;
+
+	if (!current->vbsp) {
 		return VMMERR_NOSUP;
 	}
 
 	printf("Reset %s\n", vm_get_name());
 
+	vm->resetting = true;
+
 	/*
 	 * Initialize other cpus.
 	 */
 	apic_send_init_others();
 
-	usleep(100*1000);
-
 	/*
 	 * Reset vcpu and load guest BIOS.
 	 */
 	printf("Reset cpu 0x%x\n", get_cpu_id());
 	vcpu_reset();
-	vcpu_reset_apic();
+	apic_maskall();
+
+	vm_wait_other_cpus();
+
+	if (vm->gmm.clear_mem) {
+		vm->gmm.clear_mem();
+	}
+
+	if (vm_get_id() == 0) {
+		setup_guestboot();
+	} else {
+		gbios_load();
+	}
 
-	current->vm->gmm.clear_mem();
-	gbios_load();
 	call_initfunc("resetvm");
 
+	vm_wait_other_cpus();
+	extint_reset_apic();
+	vm_wait_other_cpus();
+
+	vm->resetting = false;
+
 	return VMMERR_SUCCESS;
 }
