@@ -27,10 +27,11 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 /*
- * Copyright (c) 2010-2013 Yuichi Watanabe
+ * Copyright (c) 2010-2014 Yuichi Watanabe
  */
 
 #include <core/assert.h>
+#include <core/extint.h>
 #include <core/initfunc.h>
 #include <core/linkage.h>
 #include <core/printf.h>
@@ -231,18 +232,22 @@ do_exception (void)
 				       current->u.vt.intr.intr_info.s.err,
 				       vii.s.vector, vii.s.type, vii.s.err);
 			}
+#ifdef VT_EXCEPTION_DEBUG
 			if ((current->u.vt.intr.intr_info.v | 0x8000000)
 			    == vii.v) {
 				printf("DEBUG: Injecting the same exception twice.\n");
 			}
+#endif
 			current->u.vt.intr.intr_info.v = vii.v;
 			if (vii.s.err == INTR_INFO_ERR_VALID) {
 				asm_vmread (VMCS_VMEXIT_INTR_ERRCODE,
 					    &errc);
 				current->u.vt.intr.exception_errcode = errc;
+#ifdef VT_EXCEPTION_DEBUG
 				if (vii.s.vector == EXCEPTION_GP) {
 					printf ("DEBUG: General protection fault on CPU%d. error code 0x%lx\n", get_cpu_id(), errc);
 				}
+#endif
 			}
 			current->u.vt.intr.instruction_len = 0;
 		}
@@ -641,12 +646,14 @@ vt_exit_reason (void)
 		 * extint_check_interrupt will be inject the pending
 		 * interrupt or keep it.
 		 */
+		extint_request_check();
 		break;
 	case EXIT_REASON_INTERRUPT_WINDOW:
 		/*
 		 * extint_check_interrupt will be inject the pending
 		 * interrupt.
 		 */
+		extint_request_check();
 		break;
 	case EXIT_REASON_INVLPG:
 		do_invlpg ();
@@ -675,7 +682,15 @@ vt_exit_reason (void)
 		} else {
 			printf("Reset cpu 0x%x\n", get_cpu_id());
 			vcpu_reset();
-			vcpu_reset_apic();
+			apic_maskall();
+			if (vm_resetting()) {
+				vm_wait_other_cpus();
+				vm_wait_other_cpus();
+			}
+			extint_reset_apic();
+			if (vm_resetting()) {
+				vm_wait_other_cpus();
+			}
 		}
 		break;
 	case EXIT_REASON_XSETBV:
@@ -701,7 +716,17 @@ vt_halt (void)
 		ulong acr;
 		ulong limit;
 	} es, cs, ss, ds, fs, gs;
+	struct vt_intr_data *vid = &current->u.vt.intr;
 	ulong rflags;
+	ulong int_state;
+
+	if (vid->intr_info.s.valid == INTR_INFO_VALID_VALID) {
+		/*
+		 * RIP have been incremented by interpreter,
+		 * so that we just return.
+		 */
+		return;
+	}
 
 	asm_vmread (VMCS_GUEST_ES_SEL, &es.sel);
 	asm_vmread (VMCS_GUEST_CS_SEL, &cs.sel);
@@ -743,6 +768,21 @@ vt_halt (void)
 	asm_vmwrite (VMCS_GUEST_RFLAGS, RFLAGS_ALWAYS1_BIT | RFLAGS_IOPL_0 |
 		(rflags & RFLAGS_IF_BIT));
 	asm_vmwrite (VMCS_GUEST_ACTIVITY_STATE, VMCS_GUEST_ACTIVITY_STATE_HLT);
+
+	/*
+	 * Avoid blocking.
+	 */
+	asm_vmread (VMCS_GUEST_INTERRUPTIBILITY_STATE, &int_state);
+	int_state &= (VMCS_GUEST_INTERRUPTIBILITY_STATE_BLOCKING_BY_STI_BIT |
+		      VMCS_GUEST_INTERRUPTIBILITY_STATE_BLOCKING_BY_MOV_SS_BIT);
+	asm_vmwrite (VMCS_GUEST_INTERRUPTIBILITY_STATE, int_state);
+
+	if (rflags & RFLAGS_IF_BIT) {
+		vt_extint_pending(false);
+	} else {
+		vt_extint_pending(true);
+	}
+
 	vt_vm_run ();
 	if (false) {		/* DEBUG */
 		ulong exit_reason;
@@ -783,19 +823,15 @@ vt_mainloop (void)
 	ulong cr0, acr;
 	u64 efer;
 
-	enable_interrupt();
 	for (;;) {
 		schedule ();
 		apic_handle_wait_for_sipi_state();
 
-		disable_interrupt();
 		if (current->halt) {
 			vt_halt();
 			current->halt = false;
-			enable_interrupt();
 			continue;
 		}
-		enable_interrupt();
 
 		/* when the state is switching between real mode and
 		   protected mode, we try emulation first */
@@ -846,7 +882,7 @@ vt_mainloop (void)
 			   (sw.enable may be changed after cpu_interpreter())
 			*/
 		}
-		disable_interrupt();
+
 		extint_check_interrupt ();
 		vt_event_delivery_setup ();
 		if (current->u.vt.vr.sw.enable) {
@@ -859,7 +895,6 @@ vt_mainloop (void)
 		cpu_mmu_spt_tlbflush ();
 		vt_event_delivery_check ();
 		vt_exit_reason ();
-		enable_interrupt();
 	}
 }
 
