@@ -27,7 +27,7 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 /*
- * Copyright (c) 2010-2012 Yuichi Watanabe
+ * Copyright (c) 2010-2014 Yuichi Watanabe
  */
 
 #include "asm.h"
@@ -389,6 +389,7 @@ vt_write_control_reg (enum control_reg reg, ulong val)
 	ulong cr0_0, cr0_1;
 	ulong cr4_0, cr4_1;
 	ulong host_val;
+	ulong cr4;
 
 	switch (reg) {
 	case CONTROL_REG_CR0:
@@ -401,7 +402,12 @@ vt_write_control_reg (enum control_reg reg, ulong val)
 		if (!!(val & CR0_PG_BIT) != !!current->u.vt.vr.pg) {
 			current->u.vt.vr.pg = !!(val & CR0_PG_BIT);
 			vt_ept_pg_change(current->u.vt.vr.pg);
-			cpu_mmu_spt_updatecr3();
+
+			/*
+			 * Update actual guest cr4.
+			 */
+			vt_read_control_reg(CONTROL_REG_CR4, &cr4);
+			vt_write_control_reg(CONTROL_REG_CR4, cr4);
 		}
 		vt_msr_update_lma ();
 		if (!(current->u.vt.ept_enabled && current->u.vt.vr.pg)) {
@@ -412,7 +418,7 @@ vt_write_control_reg (enum control_reg reg, ulong val)
 		val &= cr0_1;
 		val |= cr0_0;
 		asm_vmwrite (VMCS_GUEST_CR0, val);
-		cpu_mmu_spt_updatecr3 (); /* FIXME! Do we need this? */
+		cpu_mmu_spt_updatecr3();
 		break;
 	case CONTROL_REG_CR2:
 		current->u.vt.vr.cr2 = val;
@@ -425,16 +431,23 @@ vt_write_control_reg (enum control_reg reg, ulong val)
 		cpu_mmu_spt_updatecr3();
 		break;
 	case CONTROL_REG_CR4:
-		asm_vmwrite (VMCS_CR4_READ_SHADOW, val);
-
+		/*
+		 *Update host cr4.
+		 */
 		asm_rdcr4(&host_val);
-		host_val = (host_val & ~CR4_PASSTHROUGH) | (val & CR4_PASSTHROUGH);
+		host_val = (host_val & ~CR4_PASSTHROUGH)
+			| (val & CR4_PASSTHROUGH);
 		asm_wrcr4(host_val);
-
-		asm_vmread(VMCS_HOST_CR4, &host_val);
-		host_val = (host_val & ~CR4_PASSTHROUGH) | (val & CR4_PASSTHROUGH);
 		asm_vmwrite(VMCS_HOST_CR4, host_val);
 
+		/*
+		 * Update guest cr4 shadow.
+		 */
+		asm_vmwrite (VMCS_CR4_READ_SHADOW, val);
+
+		/*
+		 * Update acutual guest cr4.
+		 */
 		asm_rdmsr (MSR_IA32_VMX_CR4_FIXED0, &cr4_0);
 		asm_rdmsr (MSR_IA32_VMX_CR4_FIXED1, &cr4_1);
 		val &= cr4_1;
@@ -446,8 +459,18 @@ vt_write_control_reg (enum control_reg reg, ulong val)
 			 */
 			val |= CR4_PAE_BIT;
 		}
+		if (!current->u.vt.vr.pg) {
+			/*
+			 * For non-paged protect mode, disable
+			 * supervisor-mode execution prevention,
+			 * because PTE_US_BIT in shadow pte is set,
+			 * and guest software runs in supervisor-mode.
+			 */
+			val &= ~CR4_SMEP_BIT;
+		}
 		asm_vmwrite(VMCS_GUEST_CR4, val);
-		cpu_mmu_spt_updatecr3 ();
+
+		cpu_mmu_spt_updatecr3();
 		break;
 	default:
 		panic ("Unknown control register.");
