@@ -218,7 +218,7 @@ pci_assign_device(struct pci_device *dev)
 	pci_data = vm_get_driver_data(pci_handle);
 	LIST2_ADD(pci_data->vm_pci_device_list, vm_pci_device_list, dev);
 
-	if (!cpu_is_bsp()) {
+	if (vm_get_id() != 0) {
 		pci_setup_assigned_device(dev);
 	}
 	pci_setup_vmresource(dev);
@@ -562,4 +562,4 @@ pci_reset_assigned_device(void)
 	}
 }
 
-BUS_RESETVM(pci_reset_assigned_device);
+BUS_RESET(pci_reset_assigned_device);
