@@ -478,3 +478,57 @@ pci_write_config_64(/* u16 seg, */ u8 bus, u8 devfn, pci_off_t offset, u64 data)
 	pci_write_config_32(bus, devfn, offset + 4, data >> 32);
 	pci_write_config_32(bus, devfn, offset, data);
 }
+
+u8
+pci_read_config8(struct pci_device *dev, pci_off_t offset)
+{
+	return pci_read_config_8(dev->bus_no, dev->devfn, offset);
+}
+
+u16
+pci_read_config16(struct pci_device *dev, pci_off_t offset)
+{
+	return pci_read_config_16(dev->bus_no, dev->devfn, offset);
+}
+
+u32
+pci_read_config32(struct pci_device *dev, pci_off_t offset)
+{
+	return pci_read_config_32(dev->bus_no, dev->devfn, offset);
+}
+
+u64
+pci_read_config64(struct pci_device *dev, pci_off_t offset)
+{
+	return pci_read_config_64(dev->bus_no, dev->devfn, offset);
+}
+
+void
+pci_write_config8(struct pci_device *dev, pci_off_t offset, u8 data)
+{
+	pci_write_config_8(dev->bus_no, dev->devfn, offset, data);
+}
+
+void
+pci_write_config16(struct pci_device *dev, pci_off_t offset, u16 data)
+{
+	pci_write_config_16(dev->bus_no, dev->devfn, offset, data);
+}
+
+void
+pci_write_config32(struct pci_device *dev, pci_off_t offset, u32 data)
+{
+	pci_write_config_32(dev->bus_no, dev->devfn, offset, data);
+}
+
+void
+pci_write_config64(struct pci_device *dev, pci_off_t offset, u64 data)
+{
+	pci_write_config_64(dev->bus_no, dev->devfn, offset, data);
+}
+
+size_t
+pci_config_size(void)
+{
+	return LIST_HEAD(pci_mmconfig_list) ? PCI_MMCONFIG_SIZE : PCI_LEGACY_CONFIG_SIZE;
+}
