@@ -38,7 +38,6 @@
  */
 
 #include <core/acpi.h>
-#include <core/cpu.h>
 #include <core/initfunc.h>
 #include <core/mm.h>
 #include <core/rm.h>
@@ -436,7 +435,7 @@ pci_setup_vm(void)
 	pci_data->config_addr.value = initial_config_addr.value;
 	LIST2_HEAD_INIT(pci_data->vm_pci_device_list);
 
-	if (cpu_is_bsp()) {
+	if (vm_get_id() == 0) {
 		pci_assign_devices_to_vm0();
 	} else {
 		pci_assign_devices();
@@ -450,4 +449,4 @@ pci_setup_vm(void)
 }
 
 BUS_INIT(pci_init);
-BUS_PASSVM(pci_setup_vm);
+BUS_SETUPVM(pci_setup_vm);
