@@ -27,7 +27,6 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#include <core/cpu.h>
 #include <core/gmm.h>
 #include <core/initfunc.h>
 #include <core/io.h>
@@ -293,7 +292,7 @@ rtc_setup_vm(void)
 
 	spinlock_init(&rtc_emu->lock);
 
-	if (cpu_is_bsp()) {
+	if (vm_get_id() == 0) {
 		rtc_emu->allow_host_access = 1;
 		rtc_emu->index
 			= (nmi_disable ? RTC_INDEX_REG_NMI_DISABLE_BIT : 0);
@@ -306,6 +305,21 @@ rtc_setup_vm(void)
 	set_iofunc(RTC_TARGET_REG_IOPORT, rtc_target_handler);
 }
 
-DRIVER_INIT(rtc_init);
-DRIVER_PASSVM(rtc_setup_vm);
+static void
+rtc_reset(void)
+{
+	u8 reg;
 
+	if (vm_get_id() != 0) {
+		return;
+	}
+	printf("Reset RTC.\n");
+	write_rtc(RTC_REGISTER_A, 0x26);
+	read_rtc(RTC_REGISTER_B, &reg);
+	write_rtc(RTC_REGISTER_B, (reg & ~0x04) | 0x02);
+	read_rtc(RTC_REGISTER_C, &reg);
+}
+
+DRIVER_INIT(rtc_init);
+DRIVER_SETUPVM(rtc_setup_vm);
+DRIVER_RESET(rtc_reset);
